#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#define SHUT_RDWR SD_BOTH
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif
#include "server.h"
#include "debug.h"
#include "ftp.h"
#include "listdir.h"
#include "md5.h"

int conn_sendmsg(struct ftpconn* conn, int code, char* str)
{
	DBG(">> %d %s", code, str);
	int len, ret;
	if(str == NULL)
		return 0;
	len = ftpmsg_format(code, str, conn->msgbuf, sizeof(conn->msgbuf)-4);
#ifdef __WIN32__
	if(conn->opts_utf8)
		gb_to_utf8(conn->msgbuf, conn->msgbuf, sizeof(conn->msgbuf)-4);
#endif
	ret = send(conn->conn_fd, conn->msgbuf, strlen(conn->msgbuf), 0);
	if(ret <= 0){
		conn->conn_end = 1;
		return 1;
	}
	return 0;
}

char* conn_readfile(struct ftpconn* conn, const char* dir, const char* filename, const char* temp){
	char path[FULLPATH_LEN+4];
	snprintf(path, FULLPATH_LEN, "%s/%s", dir, filename);
	FILE* fp = fopen(path, "rb");
	char* ptr = NULL;
	if(fp != NULL){
		unsigned long fsize;
		fseek(fp, 0, SEEK_END);
		fsize = ftell(fp);
		fseek(fp, 0, SEEK_SET);
		ptr = malloc(fsize+1);
		if( fread(ptr, fsize, 1, fp) > 0 ){
			fclose(fp);
			ptr[fsize] = '\0';
			return ptr;
		}
		fclose(fp);
	}
	ptr = malloc(strlen(temp)+1);
	strcpy(ptr, temp);
	return ptr;
}

void conn_connect(struct ftpconn* conn)
{
	if(conn->dataconn_fd == 0)
		return;
	if(conn->opts_pasv){
		int fd = accept(conn->dataconn_fd, 0, 0);
		if(fd >= 0){
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = fd;
		}else if(conn->dataconn_fd != 0){
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = 0;
		}
	}else{
		struct sockaddr_in addr;
		memset(&addr,0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(conn->port_ip);
		addr.sin_port = htons(conn->port_port);
		DBG("connecting to %s:%d", conn->port_ip, conn->port_port);
		if(connect(conn->dataconn_fd, (struct sockaddr*)&addr, sizeof(addr)) < 0){
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = 0;
		}
	}
}

void conn_command(struct ftpconn* conn, char* cmd, char* arg)
{
	int i;
	char *temp = conn->tempbuf;
	int temp_size = sizeof(conn->tempbuf)-4;
#ifdef __WIN32__
	if(*arg && conn->opts_utf8)
		utf8_to_gb(arg, arg, 1024);
#endif
	/* get command code */
	DBG("<< %s %s", cmd, arg);
	for(i = 0; i < sizeof(FtpTokenString)/sizeof(FtpTokenString[0]); i++)
		if(stricmp(FtpTokenString[i], cmd) == 0)
			break;
	
	switch(i){
	case TOK_ERROR:
		sprintf(temp, "syntax error. command %s unrecognized.", cmd);
		conn_sendmsg(conn, 500, temp);
		return;
	case TOK_BYE:
	case TOK_QUIT:{
		char* msg_goodbye = conn_readfile(conn, conn->server->server_root, 
			"xxftp.goodbye", "Goodbye!");
		conn_sendmsg(conn, 221, msg_goodbye);
		free(msg_goodbye);
		conn->conn_end = 1;
		return;
	}
	case TOK_USER:{
		int j, isdir = 0;
		if(!arg[0])
			arg = "anonymous";
		/* bugfix: username may be part of your path */
		for(j=0; arg[j]; j++){
			if(!isalpha(arg[j]) && !isdigit(arg[j]) && arg[j]!='_'){
				conn_sendmsg(conn, 530, "incorrect username.");
				return;
			}
		}
		strncpy(conn->username, arg, 63);
		sprintf(conn->home_dir, "%s/%s", conn->server->server_root, conn->username);
		strcpy(conn->cur_dir, "/");
		strcpy(conn->full_path, conn->home_dir);
		if(!file_exists(conn->home_dir, &isdir) || !isdir){
			sprintf(temp, "user %s does not exist.", conn->username);
			conn_sendmsg(conn, 530, temp);
			return;
		}
		sprintf(temp, "%s/.xxftp/password", conn->home_dir);
		if(file_exists(temp, &isdir)){
			sprintf(temp, "password required for %s.", conn->username);
			conn_sendmsg(conn, 331, temp);
		}else{
			conn_sendmsg(conn, 230, "identified.");
			conn->identified = 1;
		}
		return;
	}
	case TOK_PASS:{
		unsigned char md5_pass[16];
		md5_state_t mst;
		md5_init( &mst );
		md5_append( &mst, (md5_byte_t*)arg, strlen(arg) );
		md5_finish( &mst, (md5_byte_t*)md5_pass );
		hex_str(md5_pass, 16, arg);
		char* pass = conn_readfile(conn, conn->home_dir, ".xxftp/password", "-");
		sscanf(pass, "%s", temp);
		if(strcmp(temp, arg) == 0){
			conn_sendmsg(conn, 230, "identified.");
			conn->identified = 1;
		}else{
			conn_sendmsg(conn, 530, "not identified.");
		}
		free(pass);
		return;
	}
	default:
		if(!conn->identified){
			conn_sendmsg(conn, 530, "please login with USER and PASS.");
			return;
		}
	}
	
	conn->time_alive = time(NULL);
	
	switch(i){
	case TOK_NOOP:
		conn_sendmsg(conn, 200, "ok");
		return;
	case TOK_TYPE:
		sprintf(temp, "type set to %s.", arg);
		conn_sendmsg(conn, 200, temp);
		return;
	case TOK_SYST:
		conn_sendmsg(conn, 200, "fbos");
		return;
	case TOK_EPSV:
	case TOK_PASV:{
		if(conn->dataconn_fd){
			shutdown(conn->dataconn_fd, SHUT_RDWR);
			closesocket(conn->dataconn_fd);
		}
		conn->opts_pasv = 1;
		conn->dataconn_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		
		struct sockaddr_in addr;
		memset(&addr,0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(conn->server->server_ip);
		addr.sin_port = htons(0);
		if ( bind( conn->dataconn_fd, (struct sockaddr*)&addr, sizeof(addr) ) <0 ){
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = 0;
		}else{
			if( listen( conn->dataconn_fd, 1 ) <0 ){
				closesocket(conn->dataconn_fd);
				conn->dataconn_fd = 0;
			}else{
				int len = sizeof(addr);
				getsockname(conn->dataconn_fd, (struct sockaddr*)&addr, &len);
				int port = htons(addr.sin_port);
				
				if(i == TOK_EPSV){
					sprintf(temp, "Entering Extended Passive Mode (|||%hu|)", port);
					conn_sendmsg(conn, 229, temp);
				}else{
					unsigned char* ipnum = (char*)&conn->server->server_ipnum;
					sprintf(temp, "Entering Passive Mode (%u,%u,%u,%u,%u,%u).",
						ipnum[0], ipnum[1], ipnum[2], ipnum[3], port/256, port%256);
					conn_sendmsg(conn, 227, temp);
				}
			}
		}
		return;
	}
	case TOK_EPRT:
		conn_sendmsg(conn, 500, "implement EPRT later...");
		return;
	case TOK_PORT:{
		if(conn->dataconn_fd){
			shutdown(conn->dataconn_fd, SHUT_RDWR);
			closesocket(conn->dataconn_fd);
		}
		conn->opts_pasv = 0;
		conn->dataconn_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

		unsigned short int Num[6];
		sscanf(arg, "%hu,%hu,%hu,%hu,%hu,%hu",
			&Num[0], &Num[1], &Num[2], &Num[3], &Num[4], &Num[5]);
			
		snprintf(conn->port_ip, 30, "%hu.%hu.%hu.%hu", Num[0], Num[1], Num[2], Num[3]);
		conn->port_port = Num[4]*256+Num[5];

		conn_sendmsg(conn, 200, "PORT command successful.");
		return;
	}
	case TOK_PWD:
	case TOK_XPWD:
		sprintf(temp, "\"%s\"", conn->cur_dir);
		conn_sendmsg(conn, 257, temp);
		return;
	case TOK_LIST:
		if(conn->dataconn_fd == 0){
			conn_sendmsg(conn, 500, "PASV or PORT first.");
			return;
		}
		conn_connect(conn);
		conn_sendmsg(conn, 150, "here comes the directory listing.");
		listdir(conn, conn->full_path);
		conn_sendmsg(conn, 226, "list ok.");
		closesocket(conn->dataconn_fd);
		conn->dataconn_fd = 0;
		return;
	case TOK_REST:
		conn->file_pos = atol(arg);
		conn_sendmsg(conn, 250, "ok.");
		return;
	case TOK_FEAT:{
		static char features[] = ""
			"211-Features:\r\n"
			" EPRT\r\n"
			" EPSV\r\n"
			" MDTM\r\n"
			" PASV\r\n"
			" REST STREAM\r\n"
			" SIZE\r\n"
			" UTF8\r\n"
			"211 End\r\n";
		send(conn->conn_fd, features, sizeof(features)-1, 0);
		return;
	}
	case TOK_OPTS:
		if(stricmp(arg, "UTF8 ON")==0){
			conn->opts_utf8 = 1;
			conn_sendmsg(conn, 200, "utf8 on.");
		}else if(stricmp(arg, "UTF8 OFF")==0){
			conn->opts_utf8 = 0;
			conn_sendmsg(conn, 200, "utf8 off.");
		}else{
			conn_sendmsg(conn, 500, "unrecognized option.");
		}
		return;
	case TOK_CDUP:
		/* like cd .. */
		i = TOK_CWD;
		arg = "..";
		break;
	}
	
	if(arg[0]){
		char* newpath = conn->newpath;
		if(arg[0]=='/')
			strcpy(temp, arg);
		else
			snprintf(temp, FULLPATH_LEN, "%s/%s", conn->cur_dir, arg);
		parse_path(temp);
		snprintf(newpath, FULLPATH_LEN, "%s/%s", conn->home_dir, temp);
		switch(i){
		case TOK_CWD:{
			int isdir;
			if(file_exists(newpath, &isdir) && isdir){
				strcpy(conn->cur_dir + 1, temp);
				strcpy(conn->full_path, newpath);
				sprintf(temp, "\"%s\" ok.", conn->cur_dir);
				conn_sendmsg(conn, 250, temp);
			}else{
				conn_sendmsg(conn, 550, "failed.");
			}
			return;
		}
		case TOK_MDTM:{
			if( file_mtime(newpath, temp) == 0)
				conn_sendmsg(conn, 213, temp);
			else
				conn_sendmsg(conn, 550, "failed.");
			return;
		}
		case TOK_SIZE:{
			FILE* fp = fopen(newpath, "rb");
			if(fp){
				fseek(fp, 0, SEEK_END);
				unsigned int fsize = ftell(fp);
				fclose(fp);
				sprintf(temp, "%u", fsize);
				conn_sendmsg(conn, 231, temp);
			}else{
				conn_sendmsg(conn, 550, "failed.");
			}
			return;
		}
		case TOK_XMKD:
		case TOK_MKD:{
			if(mkdir(newpath) != 0){
				conn_sendmsg(conn, 550, "failed.");
			}else{
				conn_sendmsg(conn, 250, "ok.");
			}
			return;
		}
		case TOK_RNFR:{
			strncpy(conn->temp_path, newpath, FULLPATH_LEN);
			sprintf(temp, "rename from %s.", conn->temp_path);
			conn_sendmsg(conn, 350, temp);
			return;
		}
		case TOK_RNTO:{
			if(rename(conn->temp_path, newpath)==0){
				conn_sendmsg(conn, 250, "RNTO ok.");
			}else{
				conn_sendmsg(conn, 550, "failed.");
			}
			return;
		}
		case TOK_XRMD:
		case TOK_RMD:
			if(rmdir(newpath) == 0){
				conn_sendmsg(conn, 250, "ok.");
			}else{
				conn_sendmsg(conn, 550, "failed.");
			}
			return;
		case TOK_DELE:
			if(remove(newpath) == 0){
				conn_sendmsg(conn, 250, "ok.");
			}else{
				conn_sendmsg(conn, 550, "failed.");
			}
			return;
		case TOK_RETR:{
			if(conn->dataconn_fd == 0){
				conn_sendmsg(conn, 500, "PASV or PORT first.");
				return;
			}
			conn_connect(conn);
			FILE* fp = fopen(newpath, "rb");
			if(fp == NULL){
				conn_sendmsg(conn, 550, "failed to open file.");
				return;
			}
			fseek(fp, conn->file_pos, SEEK_SET);
			conn_sendmsg(conn, 150, "Opening BINARY mode data connection for file transfer.");
			int ret;
			while(!feof(fp)){
				conn->time_alive = time(NULL);
				ret = fread(temp, 1, temp_size, fp);
				if(ret <= 0)
					break;
				ret = send(conn->dataconn_fd, temp, ret, 0);
				if(ret <= 0)
					break;
			}
			fclose(fp);
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = 0;
			conn_sendmsg(conn, 226, "ok.");
			return;
		}
		case TOK_STOR:
		case TOK_APPE:{
			if(conn->dataconn_fd == 0){
				conn_sendmsg(conn, 500, "PASV or PORT first.");
				return;
			}
			conn_connect(conn);
			FILE* fp = fopen(newpath, (i == TOK_APPE) ? "ab" : "wb");
			if(fp == NULL){
				conn_sendmsg(conn, 550, "failed to open file.");
				return;
			}
			conn_sendmsg(conn, 150, "Opening BINARY mode data connection for file transfer.");
			int ret;
			while(!conn->conn_end){
				conn->time_alive = time(NULL);
				ret = recv(conn->dataconn_fd, temp, temp_size, 0);
				if(ret <= 0)
					break;
				ret = fwrite(temp, ret, 1, fp);
				if(ret <= 0)
					break;
			}
			fclose(fp);
			closesocket(conn->dataconn_fd);
			conn->dataconn_fd = 0;
			conn_sendmsg(conn, 226, "ok.");
			return;
		}
		}
	}

	/* implemented command */
	sprintf(temp, "%s not implemented.", cmd);
	conn_sendmsg(conn, 500, temp);
	return;
}

void* conn_start(void* p)
{
	struct ftpconn* conn = (struct ftpconn*)p;
	
	/* server is busy */
	if(conn->server->cur_ips > conn->server->max_ips - conn->server->reserved_ips || 
		conn->ip->cur_conns > conn->server->max_conns_per_ip - conn->server->reserved_ips || 
		conn->server->cur_conns > conn->server->max_conns - conn->server->reserved_ips){
		conn_sendmsg(conn, 500, "xxftp server is busy.\r\nplease try it later...");
		conn->conn_end = 1;
		closesocket(conn->conn_fd);
		return NULL;
	}
	
	char * msg_welcome = conn_readfile(conn, conn->server->server_root, 
		"xxftp.welcome", "xxftp ready.");
	conn_sendmsg(conn, 220, msg_welcome);
	free(msg_welcome);
	
	int read_cmd = 1;
	char cmd[16+4];
	char arg[1024+4];
	int pos = 0;
	for(; !conn->conn_end; ){
		char ch;
		int ret = recv(conn->conn_fd, &ch, 1, 0);
		if(ret <= 0){
			if(ret == 0){
				closesocket(conn->conn_fd);
				conn->conn_fd = 0;
			}
			conn->conn_end = 1;
			break;
		}
		if(read_cmd){
			switch(ch){
			case ' ':
				cmd[pos] = '\0';
				read_cmd = 0;
				pos = 0;
				break;
			case '\r':
				break;
			case '\n':
				cmd[pos] = arg[0] = '\0';
				pos = 0;
				conn_command(conn, cmd, arg);
				break;
			default:
				if(pos < 16)
					cmd[pos++] = toupper(ch);
			}
		}else{
			switch(ch){
			case '\r':
				break;
			case '\n':
				arg[pos] = '\0';
				read_cmd = 1;
				conn_command(conn, cmd, arg);
				pos = 0;
				break;
			default:
				if(pos < 1024)
					arg[pos++] = ch;
			}
		}
	}
	
	return NULL;
}

